@@ -45,6 +45,7 @@
 #define AUTOPHONE_MSG       "Calling your phone..."
 #define AUTODEFAULT_MSG     "Using default second-factor authentication."
 #define ENV_VAR_MSG         "Reading $DUO_PASSCODE..."
+static const char DUO_AUTH_PING_ENDPOINT[] = "/auth/v2/ping";
 
 /*
  * Finding the maximum length for the machine's hostname
@@ -246,7 +247,7 @@ _duo_seterr(struct duo_ctx *ctx, const char *fmt, ...)
 static void
 _duo_get_hostname(char *dns_fqdn, size_t dns_fqdn_size)
 {
-    struct addrinfo hints, *info, *p;
+    struct addrinfo hints, *info;
     char hostname[HOST_NAME_MAX + 1];
 
     /* gethostname may not insert a null terminator when it needs to truncate the hostname.
@@ -312,7 +313,7 @@ static duo_code_t
 _duo_json_response(struct duo_ctx *ctx) {
     JSON_Value *json;
     JSON_Object *json_obj;
-    char *p;
+    const char *p;
     int code = DUO_SERVER_ERROR;
 
     json = json_parse_string(ctx->body);
@@ -327,7 +328,7 @@ _duo_json_response(struct duo_ctx *ctx) {
         code = DUO_OK;
     }
     if (strcasecmp(p, "FAIL") == 0) {
-        char *message;
+        const char *message;
         code = json_object_get_number(json_obj, "code");
         // json_object_get_number will return 0 if "code" not found
         if (code == 0) {
@@ -343,6 +344,48 @@ _duo_json_response(struct duo_ctx *ctx) {
     return code;
 }
 
+int
+_duo_https_exchange(struct duo_ctx *ctx, const char *method, const char *uri, int msecs, int *code)
+{
+    const int max_int_digits = (241 * sizeof(int) / 100 + 1);
+    const int max_backoff_wait_secs = 32;
+    const int initial_backof_wait_secs = 1;
+    const int backoff_factor = 2;
+
+    static const char fmt[] = "Rate-limiting response received from server. Waiting for %ld seconds before retrying.";
+    char msg[(sizeof fmt) + max_int_digits];
+    int wait_secs = initial_backof_wait_secs;
+
+    while (1) {
+        HTTPScode rc;
+        time_t retry_after;
+
+        rc = https_send(ctx->https, method, uri,
+            ctx->argc, ctx->argv, ctx->ikey, ctx->skey, ctx->useragent,
+            ctx->time_offset);
+        if (rc != HTTPS_OK)
+            return rc;
+        rc = https_recv(ctx->https, code, &ctx->body, &ctx->body_len, &retry_after, msecs);
+        if (retry_after != (time_t)-1)
+            wait_secs = retry_after - time(NULL);
+
+        if (rc != HTTPS_OK || *code != 429 || wait_secs > max_backoff_wait_secs)
+            return rc;
+
+        struct timespec timeout = {
+            .tv_sec = wait_secs,
+            .tv_nsec = (float)rand() / RAND_MAX * 1000000000
+        };
+
+        snprintf(msg, sizeof msg, fmt, (long)timeout.tv_sec);
+        if (ctx->conv_status)
+            ctx->conv_status(NULL, msg);
+        nanosleep(&timeout, NULL);
+        if (retry_after == (time_t)-1)
+            wait_secs *= backoff_factor;
+    }
+}
+
 static duo_code_t
 duo_call(struct duo_ctx *ctx, const char *method, const char *uri, int msecs)
 {
@@ -361,12 +404,8 @@ duo_call(struct duo_ctx *ctx, const char *method, const char *uri, int msecs)
             }
             break;
         }
-        if ((err = https_send(ctx->https, method, uri,
-                    ctx->argc, ctx->argv, ctx->ikey, ctx->skey, ctx->useragent)) == HTTPS_OK &&
-            (err = https_recv(ctx->https, &code,
-                &ctx->body, &ctx->body_len, msecs)) == HTTPS_OK) {
+        if (_duo_https_exchange(ctx, method, uri, msecs, &code) == HTTPS_OK)
             break;
-        }
         https_close(&ctx->https);
     }
     duo_reset(ctx);
@@ -437,11 +476,8 @@ _duo_preauth(struct duo_ctx *ctx, const char *username,
     JSON_Object *response;
     _JSON_FIND_OBJECT(response, json_obj, "response", json);
     _JSON_FIND_STRING(p, response, "result", json);
-    if (p == NULL) {
-        _duo_seterr(ctx, "JSON invalid 'result': %s", p);
-        ret = DUO_SERVER_ERROR;
-    } else if (strcasecmp(p, "auth") != 0) {
-        char *output;
+    if (strcasecmp(p, "auth") != 0) {
+        const char *output;
         _JSON_FIND_STRING(output, response, "status", json);
         if (strcasecmp(p, "allow") == 0) {
                         _duo_seterr(ctx, "%s", output);
@@ -556,6 +592,7 @@ duo_login(struct duo_ctx *ctx, const char *username,
     const char *client_ip, int flags, const char *command, const int failmode)
 {
     duo_code_t ret;
+    int size;
     char buf[256];
     char *pushinfo = NULL;
     char p[256];
@@ -567,8 +604,13 @@ duo_login(struct duo_ctx *ctx, const char *username,
         return (DUO_CLIENT_ERROR);
     }
 
+    ret = duo_sync_time_offset(ctx);
+
+    if (ret == DUO_OK) {
+        ret = _duo_preauth(ctx, username, client_ip, failmode);
+    }
     /* Check preauth status */
-    if ((ret = _duo_preauth(ctx, username, client_ip, failmode)) != DUO_CONTINUE) {
+    if (ret != DUO_CONTINUE) {
         if(ret == DUO_SERVER_ERROR || ret == DUO_CONN_ERROR || ret == DUO_CLIENT_ERROR) {
             return (failmode == DUO_FAIL_SAFE) ? (DUO_FAIL_SAFE_ALLOW) : (DUO_FAIL_SECURE_DENY);
         }
@@ -599,13 +641,23 @@ duo_login(struct duo_ctx *ctx, const char *username,
     }
 
     /* Add pushinfo parameters */
-    local_ip = duo_local_ip();
-    if (asprintf(&pushinfo, "Server+IP=%s&Command=%s",
-        local_ip, command ? urlenc_encode(command) : "") < 0 ||
-        duo_add_param(ctx, "pushinfo", pushinfo) != DUO_OK) {
+    char *encoded_command = urlenc_encode(command);
+    if (encoded_command == NULL) {
         return (DUO_LIB_ERROR);
     }
+
+    local_ip = duo_local_ip();
+    size = asprintf(&pushinfo, "Server+IP=%s&Command=%s", local_ip, encoded_command);
+    free(encoded_command);
+    if (size < 0) {
+        return (DUO_LIB_ERROR);
+    }
+
+    ret = duo_add_param(ctx, "pushinfo", pushinfo);
     free(pushinfo);
+    if (ret != DUO_OK) {
+        return (DUO_LIB_ERROR);
+    }
 
     /* Try Duo authentication.  Only use the configured timeout if
      * the call is asynchronous, because async calls should return
@@ -694,9 +746,62 @@ duo_login(struct duo_ctx *ctx, const char *username,
                     result);
                 ret = DUO_SERVER_ERROR;
             }
+            _JSON_VALUE_FREE(json_new);
             break;
         }
+        _JSON_VALUE_FREE(json_new);
     }
     _JSON_VALUE_FREE(json);
     return (ret);
 }
+
+duo_code_t
+duo_sync_time_offset(struct duo_ctx *ctx) {
+    const char *body = NULL;
+    int body_len = 0;
+    long duo_time = 0;
+    long local_time = 0;
+    JSON_Value *json = NULL;
+    JSON_Object *json_obj = NULL;
+    JSON_Object *response_obj = NULL;
+    duo_code_t ret;
+
+    ctx->argc = 0; /* no params */
+    ret = duo_call(ctx, "GET", DUO_AUTH_PING_ENDPOINT, ctx->https_timeout);
+    if (ret != DUO_OK) {
+        return ret;
+    }
+    body = ctx->body;
+    body_len = ctx->body_len;
+    if (!body || body_len == 0) {
+        _duo_seterr(ctx, "No response body from server");
+        return DUO_SERVER_ERROR;
+    }
+    json = json_parse_string(body);
+    if (!json) {
+        _duo_seterr(ctx, "invalid JSON response");
+        return DUO_SERVER_ERROR;
+    }
+    json_obj = json_value_get_object(json);
+    if (!json_obj) {
+        _duo_seterr(ctx, "No JSON object in response");
+        json_value_free(json);
+        return DUO_SERVER_ERROR;
+    }
+    response_obj = json_object_get_object(json_obj, "response");
+    if (!response_obj) {
+        _duo_seterr(ctx, "JSON missing valid 'response'");
+        json_value_free(json);
+        return DUO_SERVER_ERROR;
+    }
+    duo_time = (long)json_object_get_number(response_obj, "time");
+    if (duo_time == 0) {
+        _duo_seterr(ctx, "JSON missing valid 'time'");
+        json_value_free(json);
+        return DUO_SERVER_ERROR;
+    }
+    local_time = (long)time(NULL);
+    ctx->time_offset = duo_time - local_time;
+    json_value_free(json);
+    return DUO_OK;
+}
