@@ -9,6 +9,7 @@
 
 #include "config.h"
 
+#include <arpa/inet.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/time.h>
@@ -56,6 +57,13 @@ struct https_ctx {
 
 struct https_ctx ctx;
 
+typedef enum
+{
+	CB_NONE = 0, /* First callback*/
+	CB_KEY,      /* Last was key */
+	CB_VAL       /* Last was value */
+} callback_status_t;
+
 struct https_request {
     BIO *cbio;
     BIO *body;
@@ -69,6 +77,14 @@ struct https_request {
 
     int sigpipe_ignored;
     struct sigaction old_sigpipe;
+
+    time_t retry_after;
+
+    char *value;
+    size_t value_size;
+    char* key; /* current header name */
+    size_t key_size; /* size of header name */
+    callback_status_t last_cb;
 };
 
 static int
@@ -79,15 +95,92 @@ __on_body(http_parser *p, const char *buf, size_t len)
     return (BIO_write(req->body, buf, len) != len);
 }
 
+time_t
+_parse_retry_after(const char *header_value)
+{
+    if (header_value == NULL) {
+        return (time_t)-1;
+    }
+
+    /* Try to parse as an integer (delay in seconds) */
+    char *endptr;
+    long delay_seconds = strtol(header_value, &endptr, 10);
+    if (*endptr == '\0') {
+        return time(NULL) + delay_seconds;
+    }
+
+    /* Try to parse as a date */
+    struct tm tm;
+    memset(&tm, 0, sizeof(struct tm));
+    if (strptime(header_value, "%a, %d %b %Y %H:%M:%S %Z", &tm) != NULL) {
+        return timegm(&tm);
+    }
+
+    return (time_t)-1;
+}
+
 static int
 __on_message_complete(http_parser *p)
 {
     struct https_request *req = (struct https_request *)p->data;
 
+    req->retry_after = _parse_retry_after(req->value);
+
+    free(req->value);
+    req->value = NULL;
+    req->value_size = 0;
+    free(req->key);
+    req->key = NULL;
+    req->key_size = 0;
+    req->last_cb = CB_NONE;
+
     req->done = 1;
     return (0);
 }
 
+static const char retry_after_header[] = "Retry-After";
+static const char x_retry_after_header[] = "X-Retry-After";
+
+static int
+__on_header_field(http_parser* p, const char* at, size_t length)
+{
+    struct https_request *client = p->data;
+
+    if (client->last_cb == CB_VAL)
+        client->key_size = 0;
+
+    client->key = realloc(client->key, client->key_size + length + 1);
+    memcpy(client->key + client->key_size, at, length);
+    client->key_size += length;
+    client->key[client->key_size] = 0;
+
+	client->last_cb = CB_KEY;
+
+	return 0;
+}
+
+static int
+__on_header_value(http_parser* p, const char* at, size_t length)
+{
+    struct https_request *client = p->data;
+
+    if (strcasecmp(client->key, retry_after_header) == 0
+        || strcasecmp(client->key, x_retry_after_header) == 0)
+    {
+        if (client->last_cb != CB_VAL)
+            client->value_size = 0;
+
+        client->value = realloc(client->value, client->value_size + length + 1);
+        memcpy(client->value + client->value_size, at, length);
+        client->value_size += length;
+        client->value[client->value_size] = 0;
+    }
+
+	client->last_cb = CB_VAL;
+
+	return 0;
+}
+
 static const char *
 _SSL_strerror(void)
 {
@@ -109,26 +202,47 @@ _SSL_check_server_cert(SSL *ssl, const char *hostname)
 {
     X509 *cert;
     X509_NAME *subject;
-    const GENERAL_NAME *altname;
     STACK_OF(GENERAL_NAME) *altnames;
     ASN1_STRING *tmp;
     int i, n, match = -1;
-    const char *p;
+    struct in6_addr addr;
+    int hostnametype = GEN_DNS;
+    size_t addrsize;
+
     if (SSL_get_verify_mode(ssl) == SSL_VERIFY_NONE ||
         (cert = SSL_get_peer_certificate(ssl)) == NULL) {
         return (1);
     }
+
+    /* Check if hostname is an IP address */
+    if (inet_pton(AF_INET6, hostname, &addr) == 1) {
+        hostnametype = GEN_IPADD;
+        addrsize = sizeof(struct in6_addr);
+    } else if (inet_pton(AF_INET, hostname, &addr) == 1) {
+        hostnametype = GEN_IPADD;
+        addrsize = sizeof(struct in_addr);
+    }
+
     /* Check subjectAltName */
     if ((altnames = X509_get_ext_d2i(cert, NID_subject_alt_name,
                 NULL, NULL)) != NULL) {
         n = sk_GENERAL_NAME_num(altnames);
 
         for (i = 0; i < n && match != 1; i++) {
-            altname = sk_GENERAL_NAME_value(altnames, i);
-            p = (char *)ASN1_STRING_data(altname->d.ia5);
-            if (altname->type == GEN_DNS) {
-                match = (ASN1_STRING_length(altname->d.ia5) ==
-                    strlen(p) && match_pattern(hostname, p));
+            const GENERAL_NAME *altname = sk_GENERAL_NAME_value(altnames, i);
+            if (hostnametype == altname->type) {
+                char *altptr = (char *)ASN1_STRING_data(altname->d.ia5);
+                size_t altsize = (size_t)ASN1_STRING_length(altname->d.ia5);
+
+                if (altname->type == GEN_DNS) {
+                    match = (altsize == strlen(altptr) && match_pattern(hostname, altptr));
+                } else if (altname->type == GEN_IPADD) {
+                    if ((altsize == addrsize) && !memcpy(altptr, &addr, altsize)) {
+                        match = 1;
+                    } else {
+                        match = 0;
+                    }
+                }
             }
         }
         GENERAL_NAMES_free(altnames);
@@ -144,9 +258,15 @@ _SSL_check_server_cert(SSL *ssl, const char *hostname)
             if ((tmp = X509_NAME_ENTRY_get_data(
                        X509_NAME_get_entry(subject, i))) != NULL &&
                 ASN1_STRING_type(tmp) == V_ASN1_UTF8STRING) {
-                p = (char *)ASN1_STRING_data(tmp);
-                match = (ASN1_STRING_length(tmp) ==
-                    strlen(p) && match_pattern(hostname, p));
+                const char *pattern = (char *)ASN1_STRING_data(tmp);
+                size_t patternsize = (size_t)ASN1_STRING_length(tmp);
+                if (patternsize == strlen(pattern)) {
+                    if (!strchr(pattern, '*')) {
+                        match = strcasecmp(hostname, pattern) == 0;
+                    } else if (hostnametype == GEN_DNS) {
+                        match = match_pattern(hostname, pattern);
+                    }
+                }
             }
         }
     }
@@ -314,23 +434,29 @@ _establish_connection(struct https_request * const req,
         if (connected_socket == -1) {
             continue;
         }
-        sock_flags = fcntl(connected_socket, F_GETFL, 0);
-        fcntl(connected_socket, F_SETFL, sock_flags|O_NONBLOCK);
-
-        if (connect(connected_socket, cur_res->ai_addr, cur_res->ai_addrlen) != 0 &&
-                errno != EINPROGRESS) {
-            close(connected_socket);
-            connected_socket = -1;
-            continue;
+        if ((sock_flags = fcntl(connected_socket, F_GETFL, 0)) == -1) {
+            goto fail;
         }
+
+        if (fcntl(connected_socket, F_SETFL, sock_flags|O_NONBLOCK) == -1) {
+            goto fail;
+        }
+
+        if (connect(connected_socket, cur_res->ai_addr, cur_res->ai_addrlen) != 0
+                && errno != EINPROGRESS) {
+            goto fail;
+        }
+
         socket_error = _fd_wait(connected_socket, 10000);
         if (socket_error != 1) {
-            close(connected_socket);
-            connected_socket = -1;
-            continue;
+            goto fail;
         }
+
         /* Connected! */
         break;
+    fail:
+        close(connected_socket);
+        connected_socket = -1;
     }
     cur_res = NULL;
     freeaddrinfo(res);
@@ -470,6 +596,8 @@ https_init(const char *cafile, const char *http_proxy)
     /* Set HTTP parser callbacks */
     ctx.parse_settings.on_body = __on_body;
     ctx.parse_settings.on_message_complete = __on_message_complete;
+    ctx.parse_settings.on_header_field = __on_header_field;
+    ctx.parse_settings.on_header_value = __on_header_value;
 
     return (0);
 }
@@ -655,25 +783,34 @@ _argv_to_qs(int argc, char *argv[])
 
 HTTPScode
 https_send(struct https_request *req, const char *method, const char *uri,
-    int argc, char *argv[], const char *ikey, const char *skey, const char *useragent)
+    int argc, char *argv[], const char *ikey, const char *skey, const char *useragent, long time_offset)
 {
     BIO *b64;
     HMAC_CTX *hmac;
     unsigned char MD[SHA512_DIGEST_LENGTH];
-    char *qs, *p;
+    char *qs, *p, date[128];
     int i, n, is_get;
+    time_t t;
 
     req->done = 0;
 
+    t = time(NULL) + time_offset; /* adjust time by offset */
+    strftime(date, sizeof date, "%a, %d %b %Y %T %z", localtime(&t));
+
     /* Generate query string and canonical request to sign */
-    if ((qs = _argv_to_qs(argc, argv)) == NULL ||
-        (asprintf(&p, "%s\n%s\n%s\n%s", method, req->host, uri, qs)) < 0) {
-        free(qs);
+    if ((qs = _argv_to_qs(argc, argv)) == NULL) {
         ctx.errstr = strerror(errno);
         return (HTTPS_ERR_LIB);
     }
     /* Format request */
-    if ((is_get = (strcmp(method, "GET") == 0))) {
+    is_get = (strcmp(method, "GET") == 0);
+
+    if ((asprintf(&p, "%s\n%s\n%s\n%s\n%s", date, method, req->host, uri, qs)) < 0) {
+        free(qs);
+        ctx.errstr = strerror(errno);
+        return (HTTPS_ERR_LIB);
+    }
+    if (is_get) {
         BIO_printf(req->cbio, "GET %s?%s HTTP/1.1\r\n", uri, qs);
     } else {
         BIO_printf(req->cbio, "%s %s HTTP/1.1\r\n", method, uri);
@@ -688,6 +825,7 @@ https_send(struct https_request *req, const char *method, const char *uri,
                "User-Agent: %s\r\n",
                useragent);
     /* Add signature */
+    BIO_printf(req->cbio, "X-Duo-Date: %s\r\n", date);
     BIO_puts(req->cbio, "Authorization: Basic ");
 
     if ((hmac = HMAC_CTX_new()) == NULL) {
@@ -735,7 +873,7 @@ https_send(struct https_request *req, const char *method, const char *uri,
 
 HTTPScode
 https_recv(struct https_request *req, int *code, const char **body, int *len,
-        int msecs)
+        time_t *retry_after, int msecs)
 {
     int n, err;
 
@@ -760,6 +898,8 @@ https_recv(struct https_request *req, int *code, const char **body, int *len,
     }
     *len = BIO_get_mem_data(req->body, (char **)body);
     *code = req->parser->status_code;
+    if (retry_after)
+        *retry_after = req->retry_after;
 
     return (HTTPS_OK);
 }
