@@ -10,6 +10,8 @@
 #ifndef HTTPS_H
 #define HTTPS_H
 
+#include <time.h>
+
 typedef struct https_request https_t;
 
 typedef enum {
@@ -35,7 +37,8 @@ HTTPScode https_send(
     char *params[],
     const char *ikey,
     const char *skey,
-    const char *useragent
+    const char *useragent,
+    long time_offset
 );
 
 /* Read response, return HTTP status code, set body and length if available.
@@ -47,6 +50,7 @@ HTTPScode https_recv(
     int *code,
     const char **body,
     int *length,
+    time_t *retry_after,
     int msecs
 );
 
