@@ -7,10 +7,15 @@
 # common_suites.py
 #
 
+import re
 import os
 import subprocess
+import time
 import unittest
 import sys
+import urllib.request
+import urllib.parse
+import ssl
 
 import pexpect
 from config import (
@@ -37,6 +42,8 @@ from config import (
 )
 from mockduo_context import NORMAL_CERT, SELFSIGNED_CERT, WRONGHOST_CERT, MockDuo
 
+from typing import Sequence
+
 TESTDIR = os.path.realpath(os.path.dirname(__file__))
 
 if sys.platform == "sunos5":
@@ -48,6 +55,7 @@ else:
     EOF = pexpect.EOF
     PROMPT = pexpect.EOF
 
+
 def fips_available():
     returncode = subprocess.call(
         [os.path.join(TESTDIR, "is_fips_supported.sh")],
@@ -57,6 +65,15 @@ def fips_available():
 
 
 class CommonTestCase(unittest.TestCase):
+    def assertRegexSomeline(self, result: Sequence[str], regex: str):
+        found = False
+        for line in result:
+            if re.search(regex, line):
+                found = True
+                break
+
+        self.assertTrue(found, f"Regex '{regex}' not found in any lines of {result}")
+
     def call_binary(self, *args, **kwargs):
         raise NotImplementedError
 
@@ -67,8 +84,8 @@ class CommonSuites:
         def test_missing_config_file(self):
             """Missing conf file"""
             result = self.call_binary(["-d", "-c", "/nonexistent", "true"])
-            self.assertRegex(
-                result["stderr"][0],
+            self.assertRegexSomeline(
+                result["stderr"],
                 r"Couldn't open /nonexistent: No such file or directory",
             )
 
@@ -77,8 +94,8 @@ class CommonSuites:
             with TempConfig(TESTCONF) as temp:
                 os.chmod(temp.name, 0o644)
                 result = self.call_binary(["-d", "-c", temp.name, "true"])
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     "{name} must be readable only by user '.*'".format(name=temp.name),
                 )
 
@@ -91,24 +108,24 @@ class CommonSuites:
             ]:
                 with TempConfig(config) as temp:
                     result = self.call_binary(["-d", "-c", temp.name, "true"])
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         "Missing host, ikey, or skey in {name}".format(name=temp.name),
                     )
 
         def test_corrupt_configuration_file_failsafe(self):
             with TempConfig(BAD_CORRUPT_CONF) as temp:
                 result = self.call_binary(["-d", "-c", temp.name, "true"])
-                self.assertRegex(
-                    result["stderr"][0], "Parse error in {name}".format(name=temp.name)
+                self.assertRegexSomeline(
+                    result["stderr"], "Parse error in {name}".format(name=temp.name)
                 )
                 self.assertEqual(result["returncode"], 0)
 
         def test_corrupt_configuration_file_failsecure(self):
             with TempConfig(BAD_CORRUPT_SECURE_CONF) as temp:
                 result = self.call_binary(["-d", "-c", temp.name, "true"])
-                self.assertRegex(
-                    result["stderr"][0], "Parse error in {name}".format(name=temp.name)
+                self.assertRegexSomeline(
+                    result["stderr"], "Parse error in {name}".format(name=temp.name)
                 )
                 self.assertEqual(result["returncode"], 1)
 
@@ -119,8 +136,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "whatever", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Failsafe Duo login for 'whatever'.*: Couldn't connect to .*",
                 )
 
@@ -132,8 +149,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "whatever", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0], r"Couldn't open Duo API handle for .*"
+                self.assertRegexSomeline(
+                    result["stderr"], r"Couldn't open Duo API handle for .*"
                 )
                 self.assertEqual(result["returncode"], 1)
 
@@ -149,8 +166,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", "whatever", "true"]
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"{failmode} Duo login for .* Couldn't connect to .*: certificate verify failed".format(
                             failmode=config.failmode_as_prefix()
                         ),
@@ -164,8 +181,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -181,8 +198,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", "whatever", "true"]
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"{failmode} Duo login for .*: Couldn't connect to .*: Certificate name validation failed".format(
                             failmode=config.failmode_as_prefix()
                         ),
@@ -196,8 +213,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "whatever", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Failsecure Duo login for .*: Couldn't connect to .*: Certificate name validation failed",
                 )
 
@@ -207,8 +224,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -224,8 +241,8 @@ class CommonSuites:
                         result = self.call_binary(
                             ["-d", "-c", temp.name, "-f", code, "true"]
                         )
-                        self.assertRegex(
-                            result["stderr"][0],
+                        self.assertRegexSomeline(
+                            result["stderr"],
                             r"Aborted Duo login for '{code}'.*: HTTP {code}".format(
                                 code=code
                             ),
@@ -238,8 +255,8 @@ class CommonSuites:
                         result = self.call_binary(
                             ["-d", "-c", temp.name, "-f", code, "true"]
                         )
-                        self.assertRegex(
-                            result["stderr"][0],
+                        self.assertRegexSomeline(
+                            result["stderr"],
                             r"{failmode} Duo login for '{code}'.*: HTTP {code}".format(
                                 failmode=config.failmode_as_prefix(), code=code
                             ),
@@ -252,8 +269,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", code, "true"]
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"{failmode} Duo login for '{code}'.*: Invalid ikey or skey".format(
                             failmode=config.failmode_as_prefix(), code=code
                         ),
@@ -265,8 +282,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", "whatever", "true"]
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"{failmode} Duo login for .*: Invalid ikey or skey".format(
                             failmode=config.failmode_as_prefix()
                         ),
@@ -285,8 +302,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", user, "true"]
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"{prefix} Duo login for '{user}'.*{message}".format(
                             prefix=prefix if prefix else config.failmode_as_prefix(),
                             user=user,
@@ -325,6 +342,33 @@ class CommonSuites:
                 "preauth-allow-bad_response", "JSON missing valid 'status'"
             )
 
+        def test_preauth_allow_retry_after(self):
+            start_time = time.time()
+            self.check_preauth_state(
+                "retry-after-3-preauth-allow", "preauth-allowed", prefix="Skipped"
+            )
+            execution_time = time.time() - start_time
+            # 3.x seconds executed twice
+            self.assertGreater(execution_time, 6)
+
+        def test_preauth_allow_retry_after_date(self):
+            start_time = time.time()
+            self.check_preauth_state(
+                "retry-after-date-preauth-allow", "preauth-allowed", prefix="Skipped"
+            )
+            execution_time = time.time() - start_time
+            # 3.x seconds executed twice
+            self.assertGreater(execution_time, 6)
+
+        def test_preauth_allow_rate_limited(self):
+            start_time = time.time()
+            self.check_preauth_state(
+                "rate-limited-preauth-allow", "preauth-allowed", prefix="Skipped"
+            )
+            execution_time = time.time() - start_time
+            # 1.x seconds + 2.x seconds executed twice
+            self.assertGreater(execution_time, 6)
+
     class Hosts(CommonTestCase):
         def run(self, result=None):
             with MockDuo(NORMAL_CERT):
@@ -335,8 +379,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "-h", host, "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow' from {host}: preauth-allowed".format(
                         host=host
                     ),
@@ -363,8 +407,8 @@ class CommonSuites:
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"],
                     env={},
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -374,8 +418,8 @@ class CommonSuites:
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"],
                     env={"http_proxy": "0.0.0.0"},
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -384,8 +428,8 @@ class CommonSuites:
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"],
                     env={"http_proxy": "0.0.0.0"},
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Failsafe Duo login for .*: Couldn't connect to localhost:4443:.*",
                 )
 
@@ -400,8 +444,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "hostname", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Aborted Duo login for 'hostname': correct hostname",
                 )
                 if config.get("failmode", None) == "secure":
@@ -421,8 +465,8 @@ class CommonSuites:
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"],
                     timeout=10,
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -434,8 +478,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     "FIPS mode flag specified, but OpenSSL not built with FIPS support. Failing the auth.",
                 )
 
@@ -450,8 +494,8 @@ class CommonSuites:
                     result = self.call_binary(
                         ["-d", "-c", temp.name, "-f", "auth_timeout", "true"],
                     )
-                    self.assertRegex(
-                        result["stderr"][0],
+                    self.assertRegexSomeline(
+                        result["stderr"],
                         r"Error in Duo login for 'auth_timeout': HTTP 500",
                     )
 
@@ -460,8 +504,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "failopen", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Aborted Duo login for 'failopen': correct failmode",
                 )
 
@@ -470,8 +514,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "failclosed", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Aborted Duo login for 'failclosed': correct failmode",
                 )
 
@@ -480,8 +524,8 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "enroll", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"User enrollment required",
                 )
 
@@ -509,8 +553,8 @@ class CommonSuites:
                     },
                     timeout=15,
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
@@ -522,8 +566,8 @@ class CommonSuites:
                         "SSH_CONNECTION": "1.2.3.4",
                     },
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r" Skipped Duo login for 'preauth-allow'",
                 )
 
@@ -532,13 +576,16 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "preauth-allow", "true"]
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"Skipped Duo login for 'preauth-allow'.*: preauth-allowed",
                 )
 
     class Interactive(CommonTestCase):
-        PROMPT_REGEX = ".* or option \(1-4\): $"
+        PROMPT_REGEX = ".* or option \\(1-4\\): $"
+        INITIAL_TEXT = [
+            ".*Loading config file .*",
+        ]
         PROMPT_TEXT = [
             "Duo login for foobar",
             "Choose or lose:",
@@ -549,9 +596,23 @@ class CommonSuites:
             "Passcode or option (1-4): ",
         ]
 
+        def remove_header_lines(self, s):
+            return s
+
         def assertOutputEqual(self, output, expected):
-            processed_output = [line for line in output.split("\r\n") if line != ""]
-            self.assertListEqual(processed_output, expected)
+            processed_output = [line for line in output.decode().split("\r\n") if line != ""]
+            for index, line in enumerate(processed_output):
+                if line == expected[index]:
+                    continue
+
+                if re.search(expected[index], line):
+                    continue
+
+                self.fail(
+                    "Line {index} does not match\nExpected: {expected}\nActual: {actual}".format(
+                        index=index, expected=expected[index], actual=line
+                    )
+                )
 
         def run(self, result=None):
             with MockDuo(NORMAL_CERT):
@@ -566,7 +627,8 @@ class CommonSuites:
                     process.expect(CommonSuites.Interactive.PROMPT_REGEX, timeout=10), 0
                 )
                 self.assertOutputEqual(
-                    process.match.group(0), CommonSuites.Interactive.PROMPT_TEXT
+                    self.remove_header_lines(process.match.group(0)),
+                    self.INITIAL_TEXT + CommonSuites.Interactive.PROMPT_TEXT,
                 )
                 process.sendline(b"123456")
                 self.assertEqual(
@@ -603,7 +665,7 @@ class CommonSuites:
                 self.assertOutputEqual(
                     process.before,
                     [
-                        "A"*256,
+                        "A" * 256,
                         "[3] Error in Duo login for 'foobar'",
                     ],
                 )
@@ -617,7 +679,8 @@ class CommonSuites:
                     process.expect(CommonSuites.Interactive.PROMPT_REGEX, timeout=10), 0
                 )
                 self.assertOutputEqual(
-                    process.match.group(0), CommonSuites.Interactive.PROMPT_TEXT
+                    self.remove_header_lines(process.match.group(0)),
+                    self.INITIAL_TEXT + CommonSuites.Interactive.PROMPT_TEXT,
                 )
                 process.sendline(b"3")
                 self.assertEqual(
@@ -719,7 +782,63 @@ class CommonSuites:
                 result = self.call_binary(
                     ["-d", "-c", temp.name, "-f", "bad-json", "true"],
                 )
-                self.assertRegex(
-                    result["stderr"][0],
+                self.assertRegexSomeline(
+                    result["stderr"],
                     r"invalid JSON response",
                 )
+
+    class DuoTimeSync(CommonTestCase):
+        def run(self, result=None):
+            with MockDuo(NORMAL_CERT):
+                return super(CommonSuites.DuoTimeSync, self).run(result)
+
+        def test_time_sync_success(self):
+            """Test successful time sync with /auth/v2/ping"""
+            with TempConfig(MOCKDUO_CONF) as temp:
+                result = self.call_binary([
+                    "-d", "-c", temp.name, "-f", "foobar", "true"
+                ])
+                self.assertRegexSomeline(
+                    result["stderr"],
+                    r"Successful Duo login for 'foobar'",
+                )
+                # Optionally, check for log message about time sync if available
+
+        def test_time_sync_with_skew(self):
+            """Test time sync with server time skew (offset is used)"""
+            with TempConfig(MOCKDUO_CONF) as temp:
+                # Set the server time skew to +400 seconds permanently
+                params = {"skew": 400, "mode": "permanent"}
+                query = urllib.parse.urlencode(params)
+                url = f"https://localhost:4443/mockduo/set-skew?{query}"
+                context = ssl._create_unverified_context()
+                with urllib.request.urlopen(url, context=context) as response:
+                    response.read()
+                result = self.call_binary(["-d", "-c", temp.name, "-f", "foobar", "true"])
+                self.assertRegexSomeline(
+                    result["stderr"],
+                    r"Successful Duo login for 'foobar'",
+                )
+                # Reset skew after test
+                params = {"skew": 0, "mode": "permanent"}
+                query = urllib.parse.urlencode(params)
+                url = f"https://localhost:4443/mockduo/set-skew?{query}"
+                with urllib.request.urlopen(url, context=context) as response:
+                    response.read()
+
+        def test_time_sync_with_skew_failed(self):
+            """Test time sync with server time skew (offset is used)"""
+            with TempConfig(MOCKDUO_CONF) as temp:
+                # Set skew to 400 for the next request only (ping)
+                params = {"skew": 400, "mode": "once"}
+                query = urllib.parse.urlencode(params)
+                url = f"https://localhost:4443/mockduo/set-skew?{query}"
+                context = ssl._create_unverified_context()
+                with urllib.request.urlopen(url, context=context) as response:
+                    response.read()
+                # The client will sync to +400, but the next auth will have 0 skew
+                result = self.call_binary(["-d", "-c", temp.name, "-f", "foobar", "true"])
+                self.assertRegexSomeline(
+                    result["stderr"],
+                    r"Failsafe Duo login for 'foobar': Invalid ikey or skey",
+                )
